#include "smatch.h"

static struct position get_cur_pos(void)
{
	static struct position pos;
	static struct position none;
	struct expression *expr;
	struct statement *stmt;

	expr = last_ptr_list((struct ptr_list *)big_expression_stack);
	stmt = last_ptr_list((struct ptr_list *)big_statement_stack);
	if (expr)
		pos = expr->pos;
	else if (stmt)
		pos = stmt->pos;
	else
		pos = none;
	return pos;
}

struct expression *zero_expr(void)
{
	static struct expression *zero;

	if (zero)
		return zero;

	zero = alloc_expression(get_cur_pos(), EXPR_VALUE);
	zero->value = 0;
	zero->ctype = &char_ctype;
	return zero;
}

struct expression *value_expr(long long val)
{
	struct expression *expr;

	if (!val)
		return zero_expr();

	expr = alloc_expression(get_cur_pos(), EXPR_VALUE);
	expr->value = val;
	expr->ctype = &llong_ctype;
	return expr;
}

struct expression *member_expression(struct expression *deref, int op, struct ident *member)
{
	struct expression *expr;

	expr = alloc_expression(deref->pos, EXPR_DEREF);
	expr->op = op;
	expr->deref = deref;
	expr->member = member;
	expr->member_offset = -1;
	return expr;
}

struct expression *deref_expression(struct expression *expr)
{
	struct expression *preop;

	preop = alloc_expression(expr->pos, EXPR_PREOP);
	preop->unop = expr;
	preop->op = '*';
	return preop;
}

struct expression *assign_expression(struct expression *left, struct expression *right)
{
	struct expression *expr;

	expr = alloc_expression(right->pos, EXPR_ASSIGNMENT);
	expr->op = '=';
	expr->left = left;
	expr->right = right;
	return expr;
}

struct expression *binop_expression(struct expression *left, int op, struct expression *right)
{
	struct expression *expr;

	expr = alloc_expression(right->pos, EXPR_BINOP);
	expr->op = op;
	expr->left = left;
	expr->right = right;
	return expr;
}

struct expression *array_element_expression(struct expression *array, struct expression *offset)
{
	struct expression *expr;

	expr = binop_expression(array, '+', offset);
	return deref_expression(expr);
}

struct expression *symbol_expression(struct symbol *sym)
{
	struct expression *expr;

	expr = alloc_expression(sym->pos, EXPR_SYMBOL);
	expr->symbol = sym;
	expr->symbol_name = sym->ident;
	return expr;
}

struct expression *compare_expression(struct expression *left, int op, struct expression *right)
{
	struct expression *expr;

	expr = alloc_expression(get_cur_pos(), EXPR_COMPARE);
	expr->op = op;
	expr->left = left;
	expr->right = right;
	return expr;
}

struct expression *gen_expression_from_key(struct expression *arg, const char *key)
{
	struct expression *ret = NULL;
	struct token *token, *end;
	const char *p = key;
	char buf[4095];
	char *alloc;
	size_t len;

	/* The idea is that we can parse either $0->foo or $->foo */
	if (key[0] != '$')
		goto free;
	p++;
	while (*p >= '0' && *p <= '9')
		p++;
	len = snprintf(buf, sizeof(buf), "%s\n", p);
	alloc = alloc_string(buf);

	token = tokenize_buffer(alloc, len, &end);
	if (!token)
		goto free;
	if (token_type(token) != TOKEN_STREAMBEGIN)
		goto free;
	token = token->next;

	ret = arg;
	while (token_type(token) == TOKEN_SPECIAL &&
	       token->special == SPECIAL_DEREFERENCE) {
		token = token->next;
		if (token_type(token) != TOKEN_IDENT) {
			ret = NULL;
			goto free;
		}
		ret = deref_expression(ret);
		ret = member_expression(ret, '*', token->ident);
		token = token->next;
	}

	if (token_type(token) != TOKEN_STREAMEND)
		ret = NULL;

free:
	return ret;
}

